#!/usr/bin/ksh

#
# SSS-test
# Dali Database Solutions Inc. (C) Copyright. All rights reserved.
# http://www.tns-0.com
#

# SSS-test can be run in the following two modes: 
#	primary mode to assess scalability with respect to concurrent user count or
#	auxiliary mode to obtain average service time
#
# In the primary mode, no input arguments should be provided, while the auxiliary mode requires a mandatory first
# argument of number of runs over which service time is averaged and the secondary optional argument to enable the 
# sql trace
#
# Scalability testing/SCUC (the primary mode) can be simulated under either stateless or stateful paradigms. 
#
# SSS-test supports three interarrival distributions: deterministic distribution, zero (all-at-once) distribution and
# most importantly the exponential interarriaval distribution.
#
# Number of different implementations of the exponential interarrival times (or Poisson distributed number of arrivals
# per unit of time) is provided
#
# Under the stateful paradigm, Oracle Advanced Queueing is used to exchange messages/payload and trigger the simulated
# user executions.
#
# Shell's co-process mechanism is used to record the timing information and later calculate the simulated response time.
#
#
# Refer to the "SQL Testing through Scalability with Respect to Concurrent User Count:  A Simulation Recipe" paper 
# for the details on the topics mentioned above.
#


###############################################
## Function fnc_run_test
###############################################
function fnc_stateless_run_test {
	print -n "* Start Run#=$T_RUN_ID, User#=$T_USER_ID ..."
	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	set term off feed off
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	@$QUERY_FILE
	INSERT INTO ssstest_user.ssstest_runstats_messages( run_id, iteration_id, user_id, concurrent_users)
	VALUES( $T_RUN_ID, $T_ITERATION_ID, $T_USER_ID, $P_CONCURRENT_USERS );
	commit;
	exit
EOF
}

###############################################
## Function fnc_COPput
###############################################
function fnc_COPput { print -p -- "$@"; }

###############################################
## Function fnc_COPget
###############################################
function fnc_COPget {
        while read -p -r LINE ;do
                [[ "$LINE" == "endInput" ]] && break
                print -- $LINE
        done
        # debug: ps | grep sqlplus
}

###############################################
## Function fnc_COPwrite (uses both fnc_COPput/get)
###############################################
function fnc_COPwrite {
        typeset -i T_RUN_ID_FAKE=0
	# Start the clock/record each user's start time (T_RUN_ID_FAKE), regardless of stateLESS/FULL:
	# later record each user's end time (T_RUN_ID) in fnc_statefull_go or fnc_stateless_run_test routine
	((T_RUN_ID_FAKE=T_RUN_ID+100))
	fnc_COPput "set feed off"
	fnc_COPput "INSERT INTO ssstest_user.ssstest_runstats_messages( run_id, iteration_id, user_id, concurrent_users)"
	fnc_COPput "VALUES( $T_RUN_ID_FAKE, $T_ITERATION_ID, $T_USER_ID, $P_CONCURRENT_USERS );"
	fnc_COPput "commit;"
	fnc_COPput "prompt endInput"
	fnc_COPget
}


###############################################
## Function fnc_start
###############################################
function fnc_start {
	print -n "* RUNs TOTAL = $P_RUNS\n"
#	print -n "* Verifing local connection and taking Before snapshot...\n"
	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set serveroutput on feed off

	begin 
		if ('$P_STATE' = 'F') then
			delete ssstest_user.ssstest_queue_table;
        		delete ssstest_user.ssstest_login;
		--	dbms_output.put_line( 'debug: ssstest_queue_table and ssstest_login tables have been cleaned up');
		-- else
		--	dbms_output.put_line( 'debug: no need for cleanup, state is ' || '$P_STATE');
		end if;
	end;
/
	commit;

	delete ssstest_user.ssstest_runstats_messages;
	exec pkg_ssstest_runstats.prc_snap( 0, 0, true);
	INSERT INTO ssstest_user.ssstest_runstats_messages( run_id, iteration_id, user_id, concurrent_users)
	VALUES( 0, 0, 0, 0);
	commit;
	exit 3
EOF
#	[ $? == 3 ] || { print "\nProblem with connectivity to the  DB $ORACLE_SID"; exit 1; }

	# fnc_COPstart:
	# print "* debug: Launch co-process (COP) to:"
       	# print "* 1) trigger/release a run 2) record start of login/query (in stateless) or just query (in statefull)."
	# Accepts as many messages per run as number of users in it:
	$ORACLE_HOME/bin/sqlplus -s $ORACLE_USERNAME/$ORACLE_PASSWORD |&       COPpid=$!
	# print "debug: COPpid=$COPpid"
}


###############################################
## Function fnc_take_snapshot
###############################################
function fnc_take_snapshot {
##	Used to be that next run starts right after the snapshot for the prev run is taken, not anymore...
##	print -n "* Ok take snapshot of RUN=$T_RUN_ID and start a next RUN=$(($T_RUN_ID+1)) (or quit)?"
	print -n "* Ok to take snapshot of RUN=$T_RUN_ID?"
	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set serveroutput on feed off
	var t_exit_flag number
	exec :t_exit_flag := 5			/* 5 means hold off, 10 means free to go */
	declare
		t_num_users	INT;
	begin
		select 	Max(user_id)
		into	t_num_users
		from	ssstest_user.ssstest_runstats_messages
		where	run_id = $T_RUN_ID
		    AND iteration_id = $T_ITERATION_ID;

		IF (t_num_users = $P_CONCURRENT_USERS) THEN
			-- # Take a snapshot and insert "zero" line, indicating snapshot/baseline (in ssstest_runstats_messages) for the next run
                	dbms_output.put_line( '* Snapshot '|| $T_RUN_ID ||' taken');
                	ssstest_user.pkg_ssstest_runstats.prc_snap($T_RUN_ID, $T_ITERATION_ID);
			:t_exit_flag := 10;
--			dbms_output.put_line( 'Within PL/SQL ' || :t_exit_flag);

			-- # Since it's for the next run or iteration one of these two should be incremeneted, but which one? so don't.
			-- End result is that USER#=0 will have the highest time for a RUN - time of the snapshot
			INSERT INTO ssstest_user.ssstest_runstats_messages( run_id, iteration_id, user_id, concurrent_users)
			VALUES( $T_RUN_ID, $T_ITERATION_ID, 0, $P_CONCURRENT_USERS );
			commit;
		ELSIF( '$P_LET_RUN_AWAY' = 'Y') THEN --AND $T_RUN_ID != $P_RUNS) THEN
                	dbms_output.put_line( '* Snapshot '|| $T_RUN_ID ||' taken (** although not all users have finished **)');
                	ssstest_user.pkg_ssstest_runstats.prc_snap($T_RUN_ID, $T_ITERATION_ID);
			
			INSERT INTO ssstest_user.ssstest_runstats_messages( run_id, iteration_id, user_id, concurrent_users)
			VALUES( $T_RUN_ID, $T_ITERATION_ID, 0, $P_CONCURRENT_USERS );
			commit;
		END IF;
--	exception
	end;
/
	exit :t_exit_flag
EOF
	T_FREE_TO_GO=$?
#	print "Outside of sqlplus " $T_FREE_TO_GO
}

###############################################
## Function fnc_report
###############################################
function fnc_report {
	# for the purpose of the report if distribution is Poisson, assume Exponential:
	if [ $P_DISTRIBUTION = "P" ] ;then
		P_DISTRIBUTION="M"
	fi

	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set lin 150 serveroutput on size 100000
	declare
		-- t_rampup_users	ssstest_user.ot_varray_rampup_users := ssstest_user.ot_varray_rampup_users();
		t_rampup_users	ssstest_user.ot_varray_rampup_users;
		t_sleep_times	ssstest_user.ot_varray_sleep_times;
	begin
		-- By convention, the last element of the collection is the duration:
		IF( $P_RUNS = 2) THEN
			t_rampup_users := ssstest_user.ot_varray_rampup_users( ${runs[0]}, ${runs[1]}, $P_DURATION);
		ELSIF( $P_RUNS = 3) THEN
			t_rampup_users := ssstest_user.ot_varray_rampup_users( ${runs[0]}, ${runs[1]}, ${runs[2]}, $P_DURATION);
		ELSIF( $P_RUNS = 4) THEN
			t_rampup_users := ssstest_user.ot_varray_rampup_users( ${runs[0]}, ${runs[1]}, ${runs[2]}, ${runs[3]}, $P_DURATION);
		ELSIF( $P_RUNS = 5) THEN
			t_rampup_users := ssstest_user.ot_varray_rampup_users( ${runs[0]}, ${runs[1]}, ${runs[2]}, ${runs[3]}, ${runs[4]}, $P_DURATION);
		ELSIF( $P_RUNS = 10) THEN
			t_rampup_users := ssstest_user.ot_varray_rampup_users( ${runs[0]}, ${runs[1]}, ${runs[2]}, ${runs[3]}, ${runs[4]}, ${runs[5]}, ${runs[6]}, ${runs[7]}, ${runs[8]}, ${runs[9]}, $P_DURATION);
			IF( '$P_DISTRIBUTION' = 'M') THEN
				IF( '$P_REPORT' = 'DBG') THEN
					dbms_output.put_line('debug: Number of elements in glb_sleep_times=${#glb_sleep_times[@]}');
				END IF;
-- /***************************** -- use for 'M' distribution only:
				t_sleep_times := ssstest_user.ot_varray_sleep_times ( 
					${glb_sleep_times[1]}, ${glb_sleep_times[2]}, 	-- intentionally the same element
					${glb_sleep_times[3]}, ${glb_sleep_times[4]}, 
					${glb_sleep_times[5]}, ${glb_sleep_times[6]}, 
					${glb_sleep_times[7]}, ${glb_sleep_times[8]}, 
					${glb_sleep_times[9]}, ${glb_sleep_times[10]},
					${glb_sleep_times[11]},${glb_sleep_times[12]},
					${glb_sleep_times[13]},${glb_sleep_times[14]},
					${glb_sleep_times[15]},${glb_sleep_times[16]},
					${glb_sleep_times[17]},${glb_sleep_times[18]},
					${glb_sleep_times[19]} );
-- *****************************/
null;
dbms_output.put_line('Distiribution is $P_DISTRIBUTION');
			ELSE
				t_sleep_times := ssstest_user.ot_varray_sleep_times ( 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
			END IF;
		ELSE
			dbms_output.put_line('*** Error in fnc_report ' || $P_RUNS);
		END IF;
		-- FOR i IN 0..$P_RUNS-1 LOOP
		--	dbms_output.put_line(i || ' - ' || ${runs[i]} || ', ' || ${runs[4]} ||', ' || ${runs[1]});
		--	t_rampup_users.EXTEND;
		--	t_rampup_users(i+1) := ${runs[i]};
		--END LOOP;

		ssstest_user.pkg_ssstest_runstats.prc_results( '$P_DISTRIBUTION', t_rampup_users, t_sleep_times, '$P_REPORT');
	end;
/
	exit
EOF
}

###############################################
## Function fnc_process_input_para
###############################################
function fnc_process_input_para {
	grep "^[^#]" $PARA_FILE > $TMP_PARA_FILE
	# below typeset is probably not needed (used typeset, as declare is bash specific):
	typeset -i paraLINE_NUMBER=0
	print "Test Parameters Summary ($(date)):"
	# Read 3 uncommented parameter lines (see .par file for detailed structure):
	# sed '/^#/d' $PARA_FILE | 
	while read FULL_LINE ;do
		((paraLINE_NUMBER+=1))
		# print "Line# $paraLINE_NUMBER: $FULL_LINE"
		# LINE_${paraLINE_NUMBER}=$FULL_LINE
		if [ paraLINE_NUMBER -eq 1 ] ;then
			ORACLE_SID=`print $FULL_LINE | cut -f1 -d':'`
			ORACLE_HOME=`print $FULL_LINE | cut -f2 -d':'`
			ORACLE_USERNAME=`print $FULL_LINE | cut -f3 -d':'`
			ORACLE_PASSWORD=`print $FULL_LINE | cut -f4 -d':'`
			print "$paraLINE_NUMBER: OSID=$ORACLE_SID, OHome=$ORACLE_HOME, OUsername/OPassword=$ORACLE_USERNAME/$ORACLE_PASSWORD"
			export ORACLE_SID ORACLE_HOME ORACLE_USERNAME ORACLE_PASSWORD
		elif [ paraLINE_NUMBER -eq 2 ] ;then
			P_STATE=$(print $FULL_LINE | cut -f1 -d ':')
			P_FREQUENCY_PER_DURATION=$(print $FULL_LINE | cut -f2 -d ':')
			P_DURATION=$(print $FULL_LINE | cut -f3 -d ':')
			P_ITERATIONS=$(print $FULL_LINE | cut -f4 -d ':')
			P_DISTRIBUTION=$(print $FULL_LINE | cut -f5 -d ':')
			P_LET_RUN_AWAY=$(print $FULL_LINE | cut -f6 -d ':')
			P_EXPECTED_SERVICE_TIME=$(print $FULL_LINE | cut -f7 -d ':')
			P_REPORT=$(print $FULL_LINE | cut -f8 -d ':')
			print "$paraLINE_NUMBER: STATE=$P_STATE, FPD=$P_FREQUENCY_PER_DURATION, DUR=$P_DURATION, ITER=$P_ITERATIONS, DISTR=$P_DISTRIBUTION, RUN_AWAY=$P_LET_RUN_AWAY, EST=$P_EXPECTED_SERVICE_TIME, REPORT=$P_REPORT"
		elif [ paraLINE_NUMBER -eq 3 ] ;then
			# do nothing, let next routine figure it out
			set -A runs $(tail -1 $TMP_PARA_FILE)
			# Total number of Runs:
			P_RUNS="${#runs[@]}"
			print "$paraLINE_NUMBER: Number of runs (user ramp-up details): $P_RUNS (${runs[@]})"
		elif [ paraLINE_NUMBER -eq 4 ] ;then
			print "Problem with the parameter file. Too many uncommented input lines"; exit 1;
		else
			print "Trouble with parsing parameter file - Line# $paraLINE_NUMBER: $FULL_LINE"
		fi
	done < $TMP_PARA_FILE
}



###############################################
## Function fnc_exponential_run -- invoked from Main, but in the background !
###############################################
function fnc_exponential_run {
			# FREQUENCY PER DURATION: 2nd internal loop:
			T_FREQUENCY_PER_DURATION_ID=1
			while [ T_FREQUENCY_PER_DURATION_ID -le P_FREQUENCY_PER_DURATION ] ;do
				# ((T_SLEEP_TIME=P_DURATION/P_FREQUENCY_PER_DURATION))
				# print "*** SleepTime=$T_SLEEP_TIME (calculated as Duration/FPD=$P_DURATION/$P_FREQUENCY_PER_DURATION)"
					print "*** RUN=$T_RUN_ID, ConcurrentUsers=$P_CONCURRENT_USERS, ITERATION=$T_ITERATION_ID, FPD=$T_FREQUENCY_PER_DURATION_ID"

					# Instead of a "regular" loop of users, use loop of "sleep times"
					T_USER_ID=1
					# while [ T_USER_ID -le P_CONCURRENT_USERS ]; do
					#	fnc_run_test &
					#	# T_PDF_GO_TIME=$(echo "define f(x) {if(x <=1) return(1); return (f(x-1)*x);}; scale=0; e(-1/$T_TAU)/($T_TAU^$T_USER_ID)/f($T_USER_ID)" | bc -l)
					#	# print "sleeping for $(($T_PACING+$T_PDF_GO_TIME)) sec (theta=$T_TAU, pdf=$T_PDF_GO_TIME)"
					#	((T_USER_ID+=1))
					# done

					T_SLEEP_TIME=0
					# print "P_SLEEP_TIME before loop=$P_SLEEP_TIME"
					# while [ T_SLEEP_TIME -le P_SLEEP_TIME ] ;do
					while [ T_USER_ID -le P_CONCURRENT_USERS ]; do
						# print "* $(($T_SLEEP_TIME+1)) user -- sleeps for ${sleep_times[$T_SLEEP_TIME]} sec"
						print "* $T_USER_ID user -- sleeps for ${sleep_times[$T_USER_ID]} sec"

						# Remember "globally" sleep times for the first and last users (for the report):
						# if [[ ($T_USER_ID -eq 1) || ($T_USER_ID -eq $P_CONCURRENT_USERS) ]] ;then
						#	glb_sleep_times[$T_GLB_SLEEP_TIME]=${sleep_times[$T_USER_ID]}
						#	print "glb_sleep_times[$T_GLB_SLEEP_TIME]=${glb_sleep_times[$T_GLB_SLEEP_TIME]}"
						#	((T_GLB_SLEEP_TIME+=1))
						#fi

						# ( sleep ${sleep_times[$T_SLEEP_TIME]}; fnc_run_test ) &
						########( sleep ${sleep_times[$T_USER_ID]}; fnc_run_test ) &
						# ( sleep ${sleep_times[$T_USER_ID]}; fnc_statefull_go ) &
						if [ $P_STATE = 'F' ] ;then
							( sleep ${sleep_times[$T_USER_ID]}; fnc_COPwrite; fnc_statefull_go ) &
						elif [ $P_STATE = 'L' ] ;then
							( sleep ${sleep_times[$T_USER_ID]}; fnc_COPwrite; fnc_stateless_run_test ) &
						else
							print "Wrong StateFull/StateLess flag. Value supplied is $P_STATE"; fnc_exit;
						fi

						# ((T_SLEEP_TIME+=1))
						((T_USER_ID+=1))
					done
				((T_FREQUENCY_PER_DURATION_ID+=1))
			done
}

###############################################
## Function fnc_exponential_distribution
###############################################
function fnc_exponential_distribution {
	# print "Exponential (Poisson) Distribution of arrivals is stubbed in this version. Change parameter file and re-try."
	# fnc_exit 

	typeset -i T_PACING
	typeset -i T_TAU
	typeset -i T_PDF_GO_TIME

	# Calculate average interarrival time:
	((T_TAU=P_DURATION/P_CONCURRENT_USERS))
	print "* Avg. Interarrival Time Tau=$T_TAU [sec]"

	# Check whether or not a pdf file of expoonentially distribution interarrival times exists:
	if [[ T_RUN_ID -eq 1 ]] ;then
		typeset -u T_REUSE_EXPONDIST_FLG=''
		if [ -f  ${EXPONDIST_FILE}.$P_CONCURRENT_USERS ] ;then 
			read T_REUSE_EXPONDIST_FLG?"* There appears to be an existing pdf file of exponentially distributed interarrival times. Reuse it [y/n]?"
			if [[ $T_REUSE_EXPONDIST_FLG = "Y" ]] ;then
				T_GENERATE_EXPONDIST_FLG='N'
			else
				T_GENERATE_EXPONDIST_FLG='Y'
			fi
		else
			print "* There are no pdf files of exponentially distributed interarrival times. Generating new..."
			T_GENERATE_EXPONDIST_FLG='Y'
		fi
	fi	
	if [[ $T_GENERATE_EXPONDIST_FLG = "Y" ]] ;then
		print "* Newly generated Exponentially distributed interarrival times or Poisson distributed number of arrivals (per duration)"
		fnc_generate_expondist_file
	else
		print "* Existing/Previously generated Exponentially distributed interarrival times or Poisson distributed number of arrivals"
		# debug: echo "${EXPONDIST_FILE}.$P_CONCURRENT_USERS"
		sed '/Time overrun protection/d' ${EXPONDIST_FILE}.$P_CONCURRENT_USERS
	fi

	if [ $P_DISTRIBUTION = "M" ] ;then
		fnc_exponential_main
	else
		fnc_poisson_main
	fi


}

###############################################
## Function fnc_exponential_main
###############################################
function fnc_exponential_main {
	set -A pdf_interarrival_times $(tail -1 ${EXPONDIST_FILE}.$P_CONCURRENT_USERS)
	# Total number of elements in an exponentially distributed interarrival times array:
	P_INTERARRIVAL_TIME="${#pdf_interarrival_times[@]}"
	# print "P_INTERARRIVAL_TIME=$P_INTERARRIVAL_TIME"

	##  An array of processes with their corresponding absolute sleep times (from begining of the interval, not from the prev. arrival)
	set -A sleep_times
	typeset -i T_INTERARRIVAL_TIME=0	# global subscript
	typeset -i T_SLEEP_TIME=1		# new array subscript
	typeset -i P_ARRIVALS_PER_SEC		# local subscript
	typeset -i T_ARRIVALS_PER_SEC
	typeset -i T_SUM_BEFORE_TAU=0		# number of users arriving prior to the theoretical mean interarrival time Tau
	typeset -i T_PREV_INTERARRIVAL_TIME=0	# Interarrival time from the Previous user
	typeset T_SUM_BEFORE_TAU_PERCENT
	while [ T_INTERARRIVAL_TIME -lt P_INTERARRIVAL_TIME ] ;do
		# prev login time + current interarrival time:
		((T_TIME_PLUS1=T_INTERARRIVAL_TIME+1))
		#echo "debug: T_Time_Plus1=$T_TIME_PLUS1"
		#sleep_times[$T_INTERARRIVAL_TIME]=$((${pdf_interarrival_times[$T_INTERARRIVAL_TIME]}+$T_PREV_INTERARRIVAL_TIME))
		sleep_times[$T_TIME_PLUS1]=$((${pdf_interarrival_times[$T_INTERARRIVAL_TIME]}+$T_PREV_INTERARRIVAL_TIME))

                # if [[ $T_INTERARRIVAL_TIME -le $T_TAU ]] ;then
                if [[ ${pdf_interarrival_times[$T_INTERARRIVAL_TIME]} -le $T_TAU ]] ;then
                	((T_SUM_BEFORE_TAU+=1))
                	# print "debug: Sum:  $T_SUM_BEFORE_TAU"
                fi

		T_PREV_INTERARRIVAL_TIME=${sleep_times[$T_TIME_PLUS1]}
		((T_INTERARRIVAL_TIME+=1))
	done


	typeset -i P_SLEEP_TIME=${#sleep_times[@]}
	# print "P_SLEEP_TIME=$P_SLEEP_TIME"

	((T_SUM_BEFORE_TAU_PERCENT=T_SUM_BEFORE_TAU*100/P_SLEEP_TIME))
	print "Number of interarrival times generated to be at or prior to the mean of $T_TAU sec = $T_SUM_BEFORE_TAU ($T_SUM_BEFORE_TAU_PERCENT %)"

	T_SLEEP_TIME=1
	while [ T_SLEEP_TIME -le P_SLEEP_TIME ] ;do
		# print "debug: $T_SLEEP_TIME user -- sleeps for ${sleep_times[$T_SLEEP_TIME]} sec"
        	# Remember "globally" sleep times for the first and last users (for the report):
        	if [[ ($T_SLEEP_TIME -eq 1) || ($T_SLEEP_TIME -eq $P_SLEEP_TIME) ]] ;then
        		glb_sleep_times[$T_GLB_SLEEP_TIME]=${sleep_times[$T_SLEEP_TIME]}
                	# print "debug: glb_sleep_times[$T_GLB_SLEEP_TIME]=${glb_sleep_times[$T_GLB_SLEEP_TIME]}"
                	((T_GLB_SLEEP_TIME+=1))
        	fi
		((T_SLEEP_TIME+=1))
	done
}


###############################################
## Function fnc_poisson_main
###############################################
function fnc_poisson_main {
	set -A pdf_interarrival_times $(tail -1 ${EXPONDIST_FILE}.$P_CONCURRENT_USERS)
	## Total number of elements in an array of Poisson distributed number of arrivals INCLUDING total, so substract 1:
	#### P_INTERARRIVAL_TIME=$((${#pdf_interarrival_times[@]}-1)) 	# strip out the last(total) number
	P_INTERARRIVAL_TIME=${#pdf_interarrival_times[@]}
	# print "debug: P_INTERARRIVAL_TIME=$P_INTERARRIVAL_TIME"

	## An array of processes with their corresponding absolute sleep times (from begining of the interval, not from the prev. arrival)
	set -A sleep_times
	typeset -i T_INTERARRIVAL_TIME=0	# global subscript
	typeset -i T_SLEEP_TIME=1		# new array subscript
	typeset -i P_ARRIVALS_PER_SEC		# local subscript
	typeset -i T_ARRIVALS_PER_SEC
	typeset -i T_SUM_BEFORE_TAU=0		# number of users arriving prior to the theoretical mean interarrival time Tau
	typeset -i T_PREV_INTERARRIVAL_TIME=0	# Interarrival time from the Previous user
	typeset T_SUM_BEFORE_TAU_PERCENT
	while [ T_INTERARRIVAL_TIME -lt P_INTERARRIVAL_TIME ] ;do
                P_ARRIVALS_PER_SEC=${pdf_interarrival_times[$T_INTERARRIVAL_TIME]}
                T_ARRIVALS_PER_SEC=0
		# print "debug: P_ARRIVALS_PER_SEC=$P_ARRIVALS_PER_SEC"
		# print "debug: pdf_interarrival_times=${pdf_interarrival_times[0]}"
                while [ T_ARRIVALS_PER_SEC -lt P_ARRIVALS_PER_SEC ]; do
                	sleep_times[$T_SLEEP_TIME]=$T_INTERARRIVAL_TIME
                        # print "debug: $T_SLEEP_TIME : sleep_times= ${sleep_times[$T_SLEEP_TIME]}"

                	# if [[ ${pdf_interarrival_times[$T_INTERARRIVAL_TIME]} -le $T_TAU ]] ;then
                        if [[ $T_INTERARRIVAL_TIME -le $T_TAU ]] ;then
                        	((T_SUM_BEFORE_TAU+=1))
                                # print "debug: Sum_before_TAU=$T_SUM_BEFORE_TAU"
                        fi
                        ((T_ARRIVALS_PER_SEC+=1))
                        # print "debug: sleep_time=$T_SLEEP_TIME"
                        ((T_SLEEP_TIME+=1))
                done
		((T_INTERARRIVAL_TIME+=1))
	done


	typeset -i P_SLEEP_TIME=${#sleep_times[@]}
	# print "P_SLEEP_TIME=$P_SLEEP_TIME"

	((T_SUM_BEFORE_TAU_PERCENT=T_SUM_BEFORE_TAU*100/P_SLEEP_TIME))
	print "Number of users with arrival times generated to be at or prior to the mean of $T_TAU sec = $T_SUM_BEFORE_TAU ($T_SUM_BEFORE_TAU_PERCENT %)"

	T_SLEEP_TIME=1
	while [ T_SLEEP_TIME -le P_SLEEP_TIME ] ;do
		# print "debug: $T_SLEEP_TIME user -- sleeps for ${sleep_times[$T_SLEEP_TIME]} sec"
        	# Remember "globally" sleep times for the first and last users (for the report):
        	if [[ ($T_SLEEP_TIME -eq 1) || ($T_SLEEP_TIME -eq $P_SLEEP_TIME) ]] ;then
        		glb_sleep_times[$T_GLB_SLEEP_TIME]=${sleep_times[$T_SLEEP_TIME]}
                	# print "debug: glb_sleep_times[$T_GLB_SLEEP_TIME]=${glb_sleep_times[$T_GLB_SLEEP_TIME]}"
                	((T_GLB_SLEEP_TIME+=1))
        	fi
		((T_SLEEP_TIME+=1))
	done
}

###############################################
## Function fnc_generate_expondist_file
###############################################
function fnc_generate_expondist_file {
	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set serveroutput on feed off lin 150 pages 100
	spool ${EXPONDIST_FILE}.$P_CONCURRENT_USERS

DECLARE
	t_expondists		ssstest_user.ssstest_expondist_ty;
	t_string1		VARCHAR2(4000);
	t_string2		VARCHAR2(4000);
	t_break			BOOLEAN := TRUE;
	j			INT := 1;
	t_sum			INT;
BEGIN
	WHILE (t_break) LOOP
	  IF( '$P_DISTRIBUTION' = 'M') THEN
		t_expondists := ssstest_user.fnc_ssstest_expondist( 1/$T_TAU, $P_CONCURRENT_USERS, 'pdf');
	  ELSIF( '$P_DISTRIBUTION' = 'P') THEN
		t_expondists := ssstest_user.fnc_ssstest_expondist( 1/$T_TAU, $P_CONCURRENT_USERS, 'pdf2');		-- pdf3
	  ELSE
		RAISE_APPLICATION_ERROR( -20105, 'Unknown parameter for distribution = '|| '$P_DISTRIBUTION');
	  END IF;

	  -- before 10g might result in: "ORU-10028: line length overflow, limit of 255 chars per line", so ensure it's less than 255:
	  SELECT
		 i0 || Rpad( ' ' , 4-length(i0))  ||
		 i1 || Rpad( ' ' , 4-length(i1))  ||
		 i2 || Rpad( ' ' , 4-length(i2))  ||
		 i3 || Rpad( ' ' , 4-length(i3))  ||
		 i4 || Rpad( ' ' , 4-length(i4))  ||
		 i5 || Rpad( ' ' , 4-length(i5))  ||
		 i6 || Rpad( ' ' , 4-length(i6))  ||
		 i7 || Rpad( ' ' , 4-length(i7))  ||
		 i8 || Rpad( ' ' , 4-length(i8))  ||
		 i9 || Rpad( ' ' , 4-length(i9))  ||
		 i10 || Rpad(' ', 4-length(i10)) ||
		 i11 || Rpad(' ', 4-length(i11)) ||
		 i12 || Rpad(' ', 4-length(i12)) ||
		 i13 || Rpad(' ', 4-length(i13)) ||
		 i14 || Rpad(' ', 4-length(i14)) ||
		 i15 || Rpad(' ', 4-length(i15)) ||
		 i16 || Rpad(' ', 4-length(i16)) ||
		 i17 || Rpad(' ', 4-length(i17)) ||
		 i18 || Rpad(' ', 4-length(i18)) ||
		 i19 || Rpad(' ', 4-length(i19)) ||
		 i20 || Rpad(' ', 4-length(i20)) ||
		 i21 || Rpad(' ', 4-length(i21)) ||
		 i22 || Rpad(' ', 4-length(i22)) ||
		 i23 || Rpad(' ', 4-length(i23)) ||
		 i24 || Rpad(' ', 4-length(i24)) ||
		 i25 || Rpad(' ', 4-length(i25)) ||
		 i26 || Rpad(' ', 4-length(i26)) ||
		 i27 || Rpad(' ', 4-length(i27)) ||
		 i28 || Rpad(' ', 4-length(i28)) ||
		 i29 || Rpad(' ', 4-length(i29)) ||
		 i30 || Rpad(' ', 4-length(i30)) ||
		 i31 || Rpad(' ', 4-length(i31)) 
	  INTO	t_string1
	  FROM (
		SELECT
	       	 Substr( Sum( Decode( column_value, 0, 1, 0)),1,3) i0
       		,Substr( Sum( Decode( column_value, 1, 1, 0)),1,3) i1
       		,Substr( Sum( Decode( column_value, 2, 1, 0)),1,3) i2
	        ,Substr( Sum( Decode( column_value, 3, 1, 0)),1,3) i3
	        ,Substr( Sum( Decode( column_value, 4, 1, 0)),1,3) i4
       		,Substr( Sum( Decode( column_value, 5, 1, 0)),1,3) i5
       		,Substr( Sum( Decode( column_value, 6, 1, 0)),1,3) i6
       		,Substr( Sum( Decode( column_value, 7, 1, 0)),1,3) i7
       		,Substr( Sum( Decode( column_value, 8, 1, 0)),1,3) i8
       		,Substr( Sum( Decode( column_value, 9, 1, 0)),1,3) i9
       		,Substr( Sum( Decode( column_value, 10, 1, 0)),1,3) i10
       		,Substr( Sum( Decode( column_value, 11, 1, 0)),1,3) i11
       		,Substr( Sum( Decode( column_value, 12, 1, 0)),1,3) i12
       		,Substr( Sum( Decode( column_value, 13, 1, 0)),1,3) i13
       		,Substr( Sum( Decode( column_value, 14, 1, 0)),1,3) i14
       		,Substr( Sum( Decode( column_value, 15, 1, 0)),1,3) i15
       		,Substr( Sum( Decode( column_value, 16, 1, 0)),1,3) i16
       		,Substr( Sum( Decode( column_value, 17, 1, 0)),1,3) i17
       		,Substr( Sum( Decode( column_value, 18, 1, 0)),1,3) i18
       		,Substr( Sum( Decode( column_value, 19, 1, 0)),1,3) i19
       		,Substr( Sum( Decode( column_value, 20, 1, 0)),1,3) i20
       		,Substr( Sum( Decode( column_value, 21, 1, 0)),1,3) i21
       		,Substr( Sum( Decode( column_value, 22, 1, 0)),1,3) i22
       		,Substr( Sum( Decode( column_value, 23, 1, 0)),1,3) i23
       		,Substr( Sum( Decode( column_value, 24, 1, 0)),1,3) i24
       		,Substr( Sum( Decode( column_value, 25, 1, 0)),1,3) i25
       		,Substr( Sum( Decode( column_value, 26, 1, 0)),1,3) i26
       		,Substr( Sum( Decode( column_value, 27, 1, 0)),1,3) i27
       		,Substr( Sum( Decode( column_value, 28, 1, 0)),1,3) i28
       		,Substr( Sum( Decode( column_value, 29, 1, 0)),1,3) i29
       		,Substr( Sum( Decode( column_value, 30, 1, 0)),1,3) i30
       		,Substr( Count(column_value),1,5) i31
	  FROM    ( Table( t_expondists))
	  );
	  -- dbms_output.put_line(t_string1);

	  t_string2 := ''; t_sum := 0;
	  -- FOR i IN 1..$P_CONCURRENT_USERS LOOP
	  FOR i IN 1..t_expondists.COUNT LOOP
		t_string2 := t_string2 || t_expondists(i) || ' ';	
		t_sum := t_sum + t_expondists(i);
	  END LOOP;

	
	  IF( '$P_DISTRIBUTION' = 'P') THEN 
	  	IF( length(t_string2) <= 255) THEN t_break := FALSE; 
	        ELSE 
			dbms_output.put_line('* Line length overrun protection (distr='||'$P_DISTRIBUTION'||') - '||length(t_string2)||': #'||j); 
			j := j + 1;
		END IF;
	  ELSE	-- can be only 'M'
	  	IF( ($P_EXPECTED_SERVICE_TIME != 0) AND (t_sum > ($P_DURATION-$P_EXPECTED_SERVICE_TIME)) ) THEN
			dbms_output.put_line('* Time overrun protection (distr='||'$P_DISTRIBUTION'||') - '||t_sum||': #' ||j); 
			j := j + 1;
	  	ELSE t_break := FALSE; 
		END IF;
	  END IF;
	END LOOP;

	dbms_output.put_line(CHR(10) || 'Summary of the interarrival times distribution:');
	dbms_output.put_line('0   1   2   3   4   5   6   7   8   9   10  11  12  13  14  15  16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  total');
	dbms_output.put_line('--- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ------');

	dbms_output.put_line(t_string1);
	-- IF( '$P_DISTRIBUTION' = 'M') THEN
		dbms_output.put_line(CHR(10) || 'Detailed distribution:');
		dbms_output.put_line(t_string2);
		-- before 10g might result in: "ORU-10028: line length overflow, limit of 255 chars per line"
	-- END IF;
END;
/

	spool off
	exit
EOF
}

###############################################
## Function fnc_all_at_once_distribution (Zero No-distribution)
###############################################
function fnc_all_at_once_distribution {
			# FREQUENCY PER DURATION: 2nd internal loop:
			T_FREQUENCY_PER_DURATION_ID=1
			while [ T_FREQUENCY_PER_DURATION_ID -le P_FREQUENCY_PER_DURATION ] ;do
				((T_SLEEP_TIME=P_DURATION/P_FREQUENCY_PER_DURATION))
				# print "*** SleepTime=$T_SLEEP_TIME (calculated as Duration/FPD=$P_DURATION/$P_FREQUENCY_PER_DURATION)"
					print "*** RUN=$T_RUN_ID, ConcurrentUsers=$P_CONCURRENT_USERS, ITERATION=$T_ITERATION_ID, FPD=$T_FREQUENCY_PER_DURATION_ID"
					T_USER_ID=1
					while [ T_USER_ID -le P_CONCURRENT_USERS ]; do
						fnc_COPwrite

						if [ $P_STATE = 'F' ] ;then
							fnc_statefull_go &
						elif [ $P_STATE = 'L' ] ;then
							fnc_stateless_run_test &
						else
							print "Wrong StateFull/StateLess flag. Value supplied is $P_STATE"; fnc_exit;
						fi
						((T_USER_ID+=1))
					done
				((T_FREQUENCY_PER_DURATION_ID+=1))
			done
}

###############################################
## Function fnc_deterministic_distribution
###############################################
function fnc_deterministic_distribution {
	typeset -i T_PACING
			# FREQUENCY PER DURATION: 2nd internal loop:
			T_FREQUENCY_PER_DURATION_ID=1
			while [ T_FREQUENCY_PER_DURATION_ID -le P_FREQUENCY_PER_DURATION ] ;do
				((T_SLEEP_TIME=P_DURATION/P_FREQUENCY_PER_DURATION))
				# print "*** SleepTime=$T_SLEEP_TIME (calculated as Duration/FPD=$P_DURATION/$P_FREQUENCY_PER_DURATION)"
					print "*** RUN=$T_RUN_ID, ConcurrentUsers=$P_CONCURRENT_USERS, ITERATION=$T_ITERATION_ID, FPD=$T_FREQUENCY_PER_DURATION_ID"
					# Calculate user login time/pacing:
					((T_PACING=P_DURATION/P_CONCURRENT_USERS))
					print "* PACING=$T_PACING [sec]"
					T_USER_ID=1
					while [ T_USER_ID -le P_CONCURRENT_USERS ]; do
                                                fnc_COPwrite

						if [ $P_STATE = 'F' ] ;then
							fnc_statefull_go &
						elif [ $P_STATE = 'L' ] ;then
							fnc_stateless_run_test &
						else
							print "Wrong StateFull/StateLess flag. Value supplied is $P_STATE"; fnc_exit;
						fi
						## ((T_USER_ID+=1)) ## Use these 2 lines below if sleep is governed in fnc_main, not here:
						## [[ T_USER_ID -le P_CONCURRENT_USERS ]] && { print "sleeping for $T_PACING [sec]"; sleep $T_PACING; }
						print "sleeping for $T_PACING sec"; sleep $T_PACING;
						((T_USER_ID+=1))
					done
				((T_FREQUENCY_PER_DURATION_ID+=1))
			done
}

###############################################
## Function fnc_test_oracle
###############################################
function fnc_test_oracle {
	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set feed off echo off
	whenever sqlerror exit
	-- how to suppress the error???? doesn't look nice at all
	select 1/0 from dual;
	exit 3
EOF
	[ $? != 3 ] || { print "\nConnectivity problem: DB=$ORACLE_SID User/Pass=$ORACLE_USERNAME/$ORACLE_PASSWORD"; fnc_exit; }
}

###############################################
## Function fnc_main
###############################################
function fnc_main {
	typeset -i T_RUN_ID=1
	typeset -i T_USER_ID
	typeset -i T_FREE_TO_GO
	typeset -i T_ITERATION_ID
	typeset -i T_FREQUENCY_PER_DURATION_ID

	typeset -u T_GENERATE_EXPONDIST_FLG
	set -A glb_sleep_times		# Global array to remember sleep times of first and last user for all runs
	typeset -i T_GLB_SLEEP_TIME=1	# Global subscript for the array above
        typeset -i T_ALREADY_LOGIN_USERS

	########## Stuff below is not needed any longer as now it's done in fnc_process_input_para ()
	# Line below works in Bash, but not in KSH (pdksh), but it's needed they way it is, otherwise gives wrong results:
	# e.g. wo/ extra paranthesis you get a string, not an array of elements:
	### set -A runs $(tail -1 $PARA_FILE)
	# Total number of Runs:
	### P_RUNS="${#runs[@]}"

	fnc_start

	# RUNS: the most external loop on number of users (last line in $PARA_FILE) to test - "ramp up":
	for element in $(seq 0 $((${#runs[@]} - 1))); do
		P_CONCURRENT_USERS=${runs[$element]}
		T_ITERATION_ID=1
		# ITERATIONS: 1st internal loop:
		while [ T_ITERATION_ID -le P_ITERATIONS ]; do
			if [ $P_STATE = 'F' ] ;then
                               	# debug: ps -fjH
				# Launch $P_CONCURRENT_USERS sqlplus sessions each blocked on dequeueing...
				T_USER_ID=1
				while [ T_USER_ID -le P_CONCURRENT_USERS ]; do
					# Has to be a background process, because each one gets blocked, so unless background, it waits indefinitely:
					fnc_statefull_login &
					((T_USER_ID+=1))
				done
                               	## Check whether or not all users participating in a run logged in? Cool off and continue only when everybody is in
                                ## If yes, one way is to loop until ssstest_login.count(1)=$P_CONCURRENT_USERS
                                T_ALREADY_LOGIN_USERS=0
                                while [ T_ALREADY_LOGIN_USERS -ne P_CONCURRENT_USERS ] ;do
                                        # less than elegant, but works :-(
                                        fnc_COPput "set term off"
                                        fnc_COPput "spool $TMP_ALREADY_LOGIN_USERS_FILE"
                                        # fnc_COPput "SELECT Count(*) FROM ssstest_user.ssstest_login WHERE run_id=$T_RUN_ID;"
                                        fnc_COPput "@ssstest_aux.sql $T_RUN_ID"
                                        fnc_COPput "spool off"
                                        fnc_COPput "prompt endInput"
                                        fnc_COPget
                                        T_ALREADY_LOGIN_USERS=$(tail -1 $TMP_ALREADY_LOGIN_USERS_FILE)
					if [ $P_REPORT = 'DBG' ] ;then
                                        	# print "debug: already logged-in users for run=$T_RUN_ID is $T_ALREADY_LOGIN_USERS"
						print -n "+"
					fi
                                done

			fi

			case $P_DISTRIBUTION in 
			M|P)
				print "* Exponential or Poisson Distribution"
				fnc_exponential_distribution
				fnc_exponential_run &
				sleep $P_DURATION
				# ((T_GLB_SLEEP_TIME+=2))			# The other assignment occurs in a background func
				;;
			D) 
				print "* Deterministic Distribution"
				fnc_deterministic_distribution	 	## & used only if sleep in managed here, not in fnc_deterministic_distr
				;;
			Z)
				print "* Zero (All-at-once) no-Distribution"
				fnc_all_at_once_distribution &
				sleep $P_DURATION
				;;
			*)
				print "***Error: Wrong Distribution. Correct the parameter file. Value received is $P_DISTRIBUTION"; fnc_exit;
			esac
			## print "!!! sleeping $P_DURATION in fnc_main !!!" - moved to fnc_deterministic_distr instead and left in fnc_all_distr
			## sleep $P_DURATION

			# if last run of the LET_RUN_AWAY=Y, make it N to capture in the last snapshot whatever/however is left out:
			if [ $T_RUN_ID -eq $P_RUNS ] ;then
				P_LET_RUN_AWAY="N"
			fi
			#if [ $P_LET_RUN_AWAY = "Y" && $T_RUN_ID -ne $P_RUNS ] ;then ### this one doesn't work. Needs double par: [[ ]]
			if [[ $P_LET_RUN_AWAY = "Y" ]] ;then
			# if [ $T_RUN_ID -ne $P_RUNS ] ;then
				fnc_take_snapshot
			else
				T_FREE_TO_GO=4
				## print "Going to sleep for $T_SLEEP_TIME sec"
				## sleep $T_SLEEP_TIME
				# Take a snapshot when last user is done with his query:
				while [ T_FREE_TO_GO -ne 10 ]; do
					# sleep 5
					# fnc_ok_to_run # Used to take a snapshot and start a new run at the same time. Have to separate:
					fnc_take_snapshot
					# if [ $T_FREE_TO_GO == 5 && $P_LET_RUN_AWAY -eq 'N' ] ;then stop the test (out of time)
				done
			fi
			((T_ITERATION_ID+=1))
		done
		((T_RUN_ID+=1))
	done
}

##############################################
## Function fnc_statefull_go
## Enqueue message/payload to trigger execution by a blocked dequeueing session
###############################################
function fnc_statefull_go {
        fnc_COPput "set serveroutput on feed off"
        # exec 4> /tmp/1.sql
        fnc_COPput "declare
                        enqueue_options       dbms_aq.enqueue_options_t;
                        message_properties    dbms_aq.message_properties_t;
                        message_handle        RAW(16);
                        message               ssstest_user.ot_message;
                begin
                        message := ssstest_user.ot_message( $T_RUN_ID, $T_ITERATION_ID, $T_USER_ID, localtimestamp);
                        dbms_aq.enqueue(
                                queue_name            => 'ssstest_user.ssstest_queue',
                                enqueue_options       => enqueue_options,
                                message_properties    => message_properties,
                                payload               => message,
                                msgid                 => message_handle);
                        commit;
                end;" ##  >&4
        # exec 4<&-
        fnc_COPput "/"
        # debug: fnc_COPput "list"
        fnc_COPput "prompt endInput"
        fnc_COPget
}

##############################################
## Function fnc_statefull_login
###############################################
function fnc_statefull_login {
        $ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
        connect $ORACLE_USERNAME/$ORACLE_PASSWORD
        set serveroutput on term off feed off
        var t_run_id            number
        var t_iteration_id      number
        var t_user_id           number

        -- get blocked until Poisson says ready to go:
        -- debug: begin dbms_output.put_line('Before dequeue. RUN_ID:USER_ID='||$T_RUN_ID||':'||$T_USER_ID); end;
--/
        INSERT INTO ssstest_user.ssstest_login( run_id, iteration_id, user_id) VALUES( $T_RUN_ID, $T_ITERATION_ID, $T_USER_ID);
        COMMIT;

        declare
                dequeue_options       dbms_aq.dequeue_options_t;
                message_properties    dbms_aq.message_properties_t;
                message_handle        RAW(16);
                message               ssstest_user.ot_message;
        begin
                -- no need to lock/sleep - dequeue is SYNCRONOUS, i.e. blocking
                dbms_aq.dequeue(
                        queue_name            => 'ssstest_user.ssstest_queue',
                        dequeue_options       => dequeue_options,
                        message_properties    => message_properties,
                        payload               => message,
                        msgid                 => message_handle);
                -- debug: dbms_output.enable;
                -- debug: dbms_output.put_line('Message run_id=' ||message.run_id||', user_id='||message.user_id||', timestamp='||message.lock_datetime);
                :t_run_id               := message.run_id;
                :t_iteration_id         := message.iteration_id;
                :t_user_id              := message.user_id;
        end;
/
        -- begin dbms_output.put_line('* Start Run#='|| $T_RUN_ID || ', User#=' || $T_USER_ID || '...'); end;
        begin dbms_output.put_line('* Start Run#='|| :t_run_id || ', User#=' || :t_user_id || '...'); end;
/
        @$QUERY_FILE
        INSERT INTO ssstest_user.ssstest_runstats_messages( run_id, iteration_id, user_id, concurrent_users)
        VALUES( :t_run_id, :t_iteration_id, :t_user_id, $P_CONCURRENT_USERS);

        -- Have to use message.user_id instead of $T_USER_ID otherwise messages might be dequeued by a wrong session, so this is incorrect:
        -- VALUES( $T_RUN_ID, $T_ITERATION_ID, $T_USER_ID, $P_CONCURRENT_USERS );
        commit;
        exit
EOF
}

###############################################
## Function fnc_get_service_name
###############################################
function fnc_get_service_time {
	time $ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set feed off timi off
--      alter session set events '10046 trace name context forever, level 8';
	exec pkg_ssstest_runstats.prc_service_time_snap( $T_RUNS_SERVICE_TIME, 1)	/* -- 1=begin */
	set timi on
	@$QUERY_FILE
	set timi off
--      alter session set events '10046 trace name context off';
	exec pkg_ssstest_runstats.prc_service_time_snap( $T_RUNS_SERVICE_TIME, 2)	/* -- 2=end  */
	exit
EOF
}

##############################################
## Function fnc_service_name_report
###############################################
function fnc_service_time_report {
	$ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set serveroutput on feed off
	exec pkg_ssstest_runstats.prc_service_time_results( $P_RUNS_SERVICE_TIME)
	exit
EOF
}

###############################################
## Function fnc_get_sql_trace
###############################################
function fnc_get_sql_trace {
	time $ORACLE_HOME/bin/sqlplus -s /nolog <<EOF
	connect $ORACLE_USERNAME/$ORACLE_PASSWORD
	set feed off
        alter session set events '10046 trace name context forever, level 8';
	-- alter session set tracefile_identifier='ssstest';
	set timi on
	@$QUERY_FILE
	set timi off
        alter session set events '10046 trace name context off';
	exit
EOF
}

###############################################
## Function fnc_cleanup
###############################################
function fnc_exit {
	[[ -f $TMP_PARA_FILE ]] && { rm $TMP_PARA_FILE;  }
	[[ -f $TMP_ALREADY_LOGIN_USERS_FILE ]] && { rm $TMP_ALREADY_LOGIN_USERS_FILE; }
	exit 1
}


#############################################
#############################################
## Main
#############################################
#trap 'rm $TMP_FILE' 0
#[ $# -lt 1 ] && { print "Usage: $0 <ORACLE_SID:Query File:Parameter File> "; exit 1; }

QUERY_FILE=ssstest.qry
PARA_FILE=ssstest.par
EXPONDIST_FILE=/tmp/ssstest_expondist
TMP_PARA_FILE=/tmp/ssstest.log
TMP_ALREADY_LOGIN_USERS_FILE=/tmp/ssstest_already_login_users.log

[[ -f $QUERY_FILE ]] || { print "Query File is missing. Aborting..."; exit 1; }	
[[ -f $PARA_FILE ]]  || { print "Parameters File is missing. Aborting..."; exit 1; }	
print "Dali Database Solutions Inc. (C) Copyright. All rights reserved. http://www.tns-0.com"
print "Usage: $(basename $0) <Optional: Number of runs to calculate an avg. service time> <Optional: 'trace' for sql trace>";

# Clenaup in case previous run terminated unexpectedly:
[[ -f $TMP_PARA_FILE ]] && { rm $TMP_PARA_FILE;  }
[[ -f $TMP_ALREADY_LOGIN_USERS_FILE ]] && { rm $TMP_ALREADY_LOGIN_USERS_FILE; }

# There are 2 modes to run this: primary mode - wo/ input para and auxialiary mode - to get avg. service rate (number of runs is required)
fnc_process_input_para
if [[ -z $1 ]] ;then
	## fnc_test_oracle ## work on this
	fnc_main
	fnc_report
	fnc_exit
else
	typeset -i P_RUNS_SERVICE_TIME=$1
	typeset -i T_RUNS_SERVICE_TIME=1
	## In this mode (estimating avg. service time) at least 1 command line para is needed: number of runs...
	while [ T_RUNS_SERVICE_TIME -le P_RUNS_SERVICE_TIME ] ;do
		if [[ -z $2 ]] ;then
			fnc_get_service_time
		elif [[ $2 = "trace" ]] ;then	#... and maybe even 2 para, provided that the 2nd one is 'trace'
			fnc_get_sql_trace &		# if sql trace bit is set, run it conccurently
		else
			print "In a <Give me a an avg. service time> mode, 2nd keyword can be only trace (if sql trace is also desired)"; fnc_exit;
		fi
        	## ( [ $T_RUNS_SERVICE_TIME -gt P_RUNS_SERVICE_TIME ] ) && { break; }
        	((T_RUNS_SERVICE_TIME+=1))
		## echo "T_RUNS_SERVICE_TIME=$T_RUNS_SERVICE_TIME"
	done
	if [[ -z $2 ]] ;then
		fnc_service_time_report
	fi
fi

